
from os import path

from .shader import Shader
from . import mesh

"""
Here are some sample obj files for prototyping your app!
"""

# This is an easy way to get the filepaths of some oft-used resources for displaying simple scenes.

resource_path = path.join(path.split(__file__)[0], '..', 'assets')

# Images
img_uvgrid = path.join(resource_path,'uvgrid.png')
img_colorgrid = path.join(resource_path, 'colorgrid.png')
img_white = path.join(resource_path, 'white.png')

# Meshes
obj_primitives = path.join(resource_path, 'primitives.obj')
obj_grid3D = path.join(resource_path, 'grid3D.obj')


# Shaders
shader_path = path.join(path.split(__file__)[0], '..', 'shaders')

# General, Normal Shader
genShader = Shader(open(path.join(shader_path, 'combShader.vert')).read(),
                   open(path.join(shader_path, 'combShader.frag')).read())

shadowShader = Shader(open(path.join(shader_path, 'shadowShader.vert')).read(),
                      open(path.join(shader_path, 'shadowShader.frag')).read())

aaShader = Shader(open(path.join(shader_path, 'antialiasShader.vert')).read(),
                  open(path.join(shader_path, 'antialiasShader.frag')).read())




# Meshes
def gen_fullscreen_quad():
    fullscreen_quad_data = mesh.MeshData(vertices=[-1, -1, 0, -1, 1, 0, 1, 1, 0, -1, -1, 0,  1, 1, 0, 1, -1, 0],
                                         face_indices=[0, 1, 2, 0, 2, 3],
                                         normals=[0, 0, -1] * 6,
                                         texcoords=[0, 0, 0, 1, 1, 1, 0, 0, 1, 1, 1, 0])
    fullscreen_quad = mesh.MeshLoader('Quad', fullscreen_quad_data).load_mesh()
    return fullscreen_quad


import time

def countdown_timer(total_secs, stop_iteration=False):
    """Generates an iterator that returns the time remaining from total_time (secs). Returns 0.0 when time is up, unless
    stop_iteration is True."""


    end_secs = time.time() + total_secs
    while True:
        remaining_secs = end_secs - time.time()
        if remaining_secs > 0.:
            yield remaining_secs
        else:
            if stop_iteration:
                raise StopIteration("Time is up!")
            else:
                yield 0.0  # Don't raise a StopIteration error, just return 0

def dt_timer():
    """A Generator that returns the time since last called. Never returns StopIteration."""
    last_time = time.time()
    while True:
        new_time = time.time()
        dt = new_time - last_time
        last_time = new_time
        yield dt




import numpy as np
from . import mixins
import pyglet.gl as gl

class Camera(mixins.PhysicalNode, mixins.Picklable):
    """A convenient object for controlling the scene viewing angle."""

    def __init__(self, fov_y=60., aspect=1.778, z_near=.01, z_far=4.5, x_shift=0., y_shift=0.,
                 ortho_mode=False, **kwargs):
        """
        Returns a Camera instance, which determines the world-to-screen perspective transformation.

        Args:
            fov_y (float): vertical field of view (degrees)
            aspect (float): screen width/height
            z_near (float): near clipping distance
            z_far (float): far clipping distance
            x_shift (float): horizontal lens shift
            y_shift (float): vertical lens shift
            ortho_mode (bool): Whether to use orthographic projection instead of perpective projection.

        Return:
            Camera instance
        """


        # Set intrinsic Camera attributes (must be manually applied using update() method during Scene.draw())
        self.fov_y = fov_y
        self.aspect = aspect
        self.zNear = z_near
        self.zFar = z_far
        self.x_shift = x_shift
        self.y_shift = y_shift
        self.ortho_mode = ortho_mode

        # Assign matrix attributes, then set them with the update_matrices() method.
        self.projection_matrix = None
        self.shift_matrix = None

        super(Camera, self).__init__(**kwargs)
        self.update()

    def _update_shift_matrix(self):
        """np.array: The Camera's lens-shift matrix."""
        self.shift_matrix =  np.array([[1.,           0.,           self.x_shift, 0.],
                         [0.,           1.,           self.y_shift, 0.],
                         [0.,           0.,                     1., 0.],
                         [0.,           0.,                     0., 1.]])

    def _update_projection_matrix(self):
        """np.array: The Camera's Projection Matrix.  Will be an Orthographic matrix if ortho_mode is set to True."""

        zn, zf = self.zNear, self.zFar

        # Use orthographic projection if enabled, else use a perspective projection.
        if self.ortho_mode == True:
            # replace glOrtho (https://www.opengl.org/sdk/docs/man2/xhtml/glOrtho.xml)

            persp_mat = np.array([[(2.)/(2),                  0.,         0., 0.], #  2/(right-left), x
                                  [      0., 2./(2./self.aspect),         0., 0.], #  2/(top-bottom), y
                                  [      0.,                  0., -2/(zf-zn), 0.], # -2/(zFar-zNear), z
                                  [      0.,                  0.,         0., 1.]])

        else:
            # replace gluPerspective (https://www.opengl.org/sdk/docs/man2/xhtml/gluPerspective.xml)
            ff = 1./np.tan(np.radians(self.fov_y / 2.)) # cotangent(fovy/2)
            persp_mat =  np.array([[ff/self.aspect,    0.,              0.,                 0.],
                                  [             0.,    ff,              0.,                 0.],
                                  [             0.,    0., (zf+zn)/(zn-zf), (2.*zf*zn)/(zn-zf)],
                                  [             0.,    0.,             -1.,                 0.]])
            persp_mat = np.dot(persp_mat, self.shift_matrix)  # Apply lens shift

        self.projection_matrix = persp_mat

    def update(self):
        super(Camera, self).update()
        self._update_shift_matrix()
        self._update_projection_matrix()

    def reset_aspect(self):
        """Gets the viewport size, and matches the camera aspect ratio to it."""
        viewport_size = (gl.GLint * 4)()
        gl.glGetIntegerv(gl.GL_VIEWPORT, viewport_size)
        self.aspect = float(viewport_size[2]) / viewport_size[3]





from pyglet import gl
from ctypes import byref, create_string_buffer, c_char, c_char_p, c_int, c_float, cast, pointer, POINTER
import numpy as np
from .utils import gl as ugl

class Uniform(object):

    _sendfuns = {'f': [gl.glUniform1f, gl.glUniform2f, gl.glUniform3f, gl.glUniform4f],
                'i':   [gl.glUniform1i, gl.glUniform2i, gl.glUniform3i, gl.glUniform4i]
                }

    def __init__(self, name, *vals):
        """A fixed-length, fixed-type array with a pre-assigned glUniform function for quick shader data sending."""
        self.name = name.encode('ascii')
        assert len(vals) > 0 and len(vals) <= 4
        self._value = np.array(vals)  # A semi-mutable array, in that its length can't be modified.
        self.sendfun = Uniform._sendfuns[self._value.dtype.kind][len(self._value) - 1]
        self.loc = None

    def __repr__(self):
        return '{}{}'.format(self.name, tuple(self.value.tolist()))

    def __getitem__(self, item):
        return self.value[item]

    def __setitem__(self, item, value):
        self.value[item] = value

    @property
    def value(self):
        return self._value

    def send_to(self, shader):
        """Sends uniform to a currently-bound shader, returning its location (-1 means not sent)"""
        # glGetUniformLocation only needs to be called once, when the shader is linked.  Not a big performance boost, though.
        if type(self.loc) == type(None):
            self.loc = gl.glGetUniformLocation(shader.id, self.name)

        self.sendfun(self.loc, *self.value)
        return self.loc


    @classmethod
    def from_dict(cls, data_dict):
        """A factory function that can build multiple uniforms from a name: val dictionary"""
        # Change all kwarg values to a sequence, to be put into Uniform
        for key, val in list(data_dict.items()):
            if not isinstance(val, (list, tuple)):
                data_dict[key] = [val]

        return [cls(key, *val) for key, val in list(data_dict.items())]


class UniformCollection(object):

    def __init__(self, uniform_dict={}):
        self._uniforms = {}
        for key, value in uniform_dict.items():
            self[key] = value

    def __setitem__(self, key, value):

        if type(value) == Uniform:
            self._uniforms[key] = value

        elif type(value) != Uniform:
            # try:
            # if key in self._uniforms:
            try:
                self._uniforms[key].value[:] = value
            except KeyError:
                try:
                    iter(value)
                    self._uniforms[key] = Uniform(key, *value)
                except TypeError:
                    self._uniforms[key] = Uniform(key, value)
            except:
                raise TypeError("Attempt to add uniform {} to UniformCollection failed.".format(key))

    def __getitem__(self, key):
        return self._uniforms[key]

    def __str__(self):
        return "UniformCollection: {}".format(self._uniforms)

    def __repr__(self):
        return self.__str__()

    def send_to(self, shader):
        for uniform in self._uniforms.values():
            uniform.send_to(shader)


#
# Copyright Tristam Macdonald 2008
#
# Distributed under the Boost Software License, Version 1.0
# (see http://www.boost.org/LICENSE_1_0.txt)
#


class Shader(ugl.BindingContextMixin, ugl.BindNoTargetMixin):

    bindfun = gl.glUseProgram
    uniformf_funs = (gl.glUniform1f, gl.glUniform2f, gl.glUniform3f, gl.glUniform4f)
    uniformi_funs = (gl.glUniform1i, gl.glUniform2i, gl.glUniform3i, gl.glUniform4i)

    def __init__(self, vert='', frag='', geom=''):
        """
        GLSL Shader program object for rendering in OpenGL.

        Args:
          - vert (str): The vertex shader program  string
          - frag (str): The fragment shader program string
          - geom (str): The geometry shader program
        """
        self.id = gl.glCreateProgram()  # create the program handle
 
        # create the vertex, fragment, and geometry shaders
        self.createShader(vert, gl.GL_VERTEX_SHADER)
        self.createShader(frag, gl.GL_FRAGMENT_SHADER)
        if geom:
            self.createShader(geom, gl.GL_GEOMETRY_SHADER_EXT)
 
        # attempt to link the program
        self.link()
 
    def createShader(self, strings, shadertype):
 
        # create the shader handle
        shader = gl.glCreateShader(shadertype)
 
        # convert the source strings into a ctypes pointer-to-char array, and upload them
        # this is deep, dark, dangerous black magick - don't try stuff like this at home!
        strings = tuple(s.encode('ascii') for s in strings)  # Nick added, for python3
        src = (c_char_p * len(strings))(*strings)
        gl.glShaderSource(shader, len(strings), cast(pointer(src), POINTER(POINTER(c_char))), None)
        # compile the shader
        gl.glCompileShader(shader)

        # retrieve the compile status
        compile_success = c_int(0)
        gl.glGetShaderiv(shader, gl.GL_COMPILE_STATUS, byref(compile_success))
 
        # if compilation failed, print the log
        if compile_success:
            gl.glAttachShader(self.id, shader)
        else:
            gl.glGetShaderiv(shader, gl.GL_INFO_LOG_LENGTH, byref(compile_success))  # retrieve the log length
            buffer = create_string_buffer(compile_success.value)  # create a buffer for the log
            gl.glGetShaderInfoLog(shader, compile_success, None, buffer)  # retrieve the log text
            print(buffer.value)  # print the log to the console
 
    def link(self):
        """link the program"""
        gl.glLinkProgram(self.id)

        # Check if linking was successful.  If not, print the log.
        link_status = c_int(0)
        gl.glGetProgramiv(self.id, gl.GL_LINK_STATUS, byref(link_status))
        if not link_status:
            gl.glGetProgramiv(self.id, gl.GL_INFO_LOG_LENGTH, byref(link_status))  # retrieve the log length
            buffer = create_string_buffer(link_status.value)  # create a buffer for the log
            gl.glGetProgramInfoLog(self.id, link_status, None, buffer)  # retrieve the log text
            print(buffer.value)  # print the log to the console

    def get_uniform_location(self, name):
        return gl.glGetUniformLocation(self.id, name.encode('ascii'))

    def uniformf(self, name, *vals):
        """Send data as a float uniform, named 'name'.  Shader must be already bound."""
        self.uniformf_funs[len(vals)-1](self.get_uniform_location(name), *vals)

    def uniformi(self, name, *vals):
        """Send data as an integer uniform, named 'name'.  Shader must be already bound."""
        self.uniformi_funs[len(vals)-1](self.get_uniform_location(name), *vals)

    def uniform_matrixf(self, name, mat, loc=None):
        """Send 4x4 NumPy matrix data as a uniform to the shader, named 'name'. Shader must be already bound."""
        # obtain the uniform location
        if not loc:
            loc = self.get_uniform_location(name)
        gl.glUniformMatrix4fv(loc, 1, False, (c_float * 16)(*mat))  # uplaod the 4x4 floating point matrix





import pyglet.gl as gl
from ctypes import byref
import contextlib
import numpy as np

def create_opengl_object(gl_gen_function, n=1):
    """Returns int pointing to an OpenGL texture"""
    handle = gl.GLuint(1)
    gl_gen_function(n, byref(handle))  # Create n Empty Objects
    if n > 1:
        return [handle.value + el for el in range(n)]  # Return list of handle values
    else:
        return handle.value  # Return handle value


@contextlib.contextmanager
def enable_states(gl_states):
    """Context Manager that calls glEnable and glDisable on a list of gl states."""
    for state in gl_states:
        gl.glEnable(state)
    yield
    for state in gl_states:
        gl.glDisable(state)



def vec(floatlist, newtype='float'):
        """ Makes GLfloat or GLuint vector containing float or uint args.
        By default, newtype is 'float', but can be set to 'int' to make
        uint list. """

        if 'float' in newtype:
            return (gl.GLfloat * len(floatlist))(*list(floatlist))
        elif 'int' in newtype:
            return (gl.GLuint * len(floatlist))(*list(floatlist))

class GlGenMixin(object):

    genfun = None

    def __init__(self, *args, **kwargs):
        super(GlGenMixin, self).__init__(*args, **kwargs)
        self.id = create_opengl_object(self.genfun)



class BindingContextMixin(object):
    """Mixin that calls self.bind() and self.unbind() when used in a context manager."""

    def __enter__(self):

        self.bind()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.unbind()


class BindTargetMixin(object):
    """Mixin that speifices a bind() and unbind() interface by taking advantage of the OpenGL bind format:
    bind: bindfun(target, id)
    unbind: bindfun(target, 0)
    """

    bindfun = None

    def bind(self):
        self.bindfun(self.target, self.id)

    @classmethod
    def unbind(cls):
        cls.bindfun(cls.target, 0)


class BindNoTargetMixin(BindTargetMixin):
    """Same as BindTargetMixin, but for bind functions that don't have a specified target."""

    def bind(self):
        self.bindfun(self.id)

    @classmethod
    def unbind(cls):
        cls.bindfun(0)


class VAO(GlGenMixin, BindingContextMixin, BindNoTargetMixin):

    genfun = gl.glGenVertexArrays
    bindfun = gl.glBindVertexArray

    def __init__(self, indices=None, **kwargs):
        """
        OpenGL Vertex Array Object.  Sends array data in a Vertex Buffer to the GPU.  This data can be accessed in
        the vertex shader using the 'layout(location = N)' header line, where N = the index of the array given the VAO.

        Example:  VAO(vertices, normals, texcoords):

        Fragshader:
        layout(location = 0) in vec3 vertexCoord;
        layout(location = 1) in vec2 texCoord;
        layout(location = 2) in vec3 normalCoord;
        """

        # Create Vertex Array Object and Bind it
        super(VAO, self).__init__(**kwargs)
        self.n_verts = None

        self.drawfun = self._draw_arrays
        self.__element_array_buffer = None
        self.element_array_buffer = indices

    @property
    def element_array_buffer(self):
        return self.__element_array_buffer

    @element_array_buffer.setter
    def element_array_buffer(self, value):
        assert isinstance(value, (np.ndarray, type(None)))
        if isinstance(value, np.ndarray):
            self.__element_array_buffer = ElementArrayBuffer(value)
            self.drawfun = self._draw_elements
        else:
            self.__element_array_buffer = None
            self.drawfun = self._draw_arrays

    def _draw_arrays(self, mode=gl.GL_TRIANGLES):
        gl.glDrawArrays(mode, 0, self.n_verts)

    def _draw_elements(self, mode=gl.GL_TRIANGLES):
        with self.element_array_buffer as el_array:
            gl.glDrawElements(mode, el_array.ndarray.shape[0],
                              gl.GL_UNSIGNED_INT, 0)

    def assign_vertex_attrib_location(self, vbo, location):
        """Load data into a vbo"""
        with vbo:
            if self.n_verts:
                assert vbo.ndarray.shape[0] == self.n_verts
            else:
                self.n_verts = vbo.ndarray.shape[0]

            # vbo.buffer_data()
            gl.glVertexAttribPointer(location, vbo.ndarray.shape[1], gl.GL_FLOAT, gl.GL_FALSE, 0, 0)
            gl.glEnableVertexAttribArray(location)

    def draw(self, mode=gl.GL_TRIANGLES):
        self.drawfun(mode)


class VBO(GlGenMixin, BindingContextMixin, BindTargetMixin):

    genfun = gl.glGenBuffers
    target = gl.GL_ARRAY_BUFFER
    bindfun = gl.glBindBuffer

    def __init__(self, ndarray, *args, **kwargs):
        super(VBO, self).__init__(*args, **kwargs)
        self.ndarray = ndarray
        self._buffer_data()

    def _buffer_data(self):
        with self:
            gl.glBufferData(self.target, 4 * self.ndarray.size, vec(self.ndarray.ravel()), gl.GL_STATIC_DRAW)


class ElementArrayBuffer(VBO):

    target = gl.GL_ELEMENT_ARRAY_BUFFER

    def __init__(self, *args, **kwargs):
        super(ElementArrayBuffer, self).__init__(*args, **kwargs)

    def _buffer_data(self):
        with self:
            gl.glBufferData(self.target, 4 * self.ndarray.size, vec(self.ndarray.ravel(), 'int'), gl.GL_STATIC_DRAW)


def setpriority(pid=None,priority=1):
    
    """ Set The Priority of a Windows Process.  Priority is a value between 0-5 where
        2 is normal priority.  Default sets the priority of the current
        python process but can take any valid process ID. """
        
    import win32api,win32process,win32con
    
    priorityclasses = [win32process.IDLE_PRIORITY_CLASS,
                       win32process.BELOW_NORMAL_PRIORITY_CLASS,
                       win32process.NORMAL_PRIORITY_CLASS,
                       win32process.ABOVE_NORMAL_PRIORITY_CLASS,
                       win32process.HIGH_PRIORITY_CLASS,
                       win32process.REALTIME_PRIORITY_CLASS]
    if pid == None:
        pid = win32api.GetCurrentProcessId()
    handle = win32api.OpenProcess(win32con.PROCESS_ALL_ACCESS, True, pid)
    win32process.SetPriorityClass(handle, priorityclasses[priority])	



import unittest
from ratcave.wavefront import WavefrontReader


class TestWavefrontReader(unittest.TestCase):
    """
    Run tests from main project folder.
    """

    def setUp(self):
        pass

    def tearDown(self):
        pass

    def test_parse_mtl(self):
        self.reader = WavefrontReader('ratcave/assets/primitives')

        assert(len(self.reader.materials) == 1)




from .utils import gl as ugl
from . import texture as tex

import pyglet.gl as gl


class FBO(ugl.GlGenMixin, ugl.BindingContextMixin):

    genfun = gl.glGenFramebuffersEXT
    target = gl.GL_FRAMEBUFFER_EXT

    def __init__(self, texture, *args, **kwargs):
        """A Framebuffer object, which when bound redirects draws to its texture.  This is useful for deferred rendering."""

        super(FBO, self).__init__(*args, **kwargs)
        self._old_viewport_size = (gl.GLint * 4)()
        self.texture = texture
        self.renderbuffer = tex.RenderBuffer(texture.width, texture.height) if not isinstance(texture, tex.DepthTexture) else None

        with self: #, self.texture:  # TODO: Figure out whether texture should also be bound here.

            # Attach the textures to the FBO
            for texture in [self.texture, self.renderbuffer] if self.renderbuffer else [self.texture]:
                texture.attach_to_fbo()

            # Set Draw and Read locations for the FBO (currently, just turn it off if not doing any color stuff)
            if isinstance(texture, tex.DepthTexture):
                gl.glDrawBuffer(gl.GL_NONE)  # No color in this buffer
                gl.glReadBuffer(gl.GL_NONE)

        # check FBO status (warning appears for debugging)
        FBOstatus = gl.glCheckFramebufferStatusEXT(gl.GL_FRAMEBUFFER_EXT)
        if FBOstatus != gl.GL_FRAMEBUFFER_COMPLETE_EXT:
            raise BufferError("GL_FRAMEBUFFER_COMPLETE failed, CANNOT use FBO.\n{0}\n".format(FBOstatus))

    def bind(self):
        """Bind the FBO.  Anything drawn afterward will be stored in the FBO's texture."""
        # This is called simply to deal with anything that might be currently bound (for example, Pyglet objects),
        gl.glBindTexture(gl.GL_TEXTURE_2D, 0)

        # Store current viewport size for later
        gl.glGetIntegerv(gl.GL_VIEWPORT, self._old_viewport_size)

        # Bind the FBO, and change the viewport to fit its texture.
        gl.glBindFramebufferEXT(gl.GL_FRAMEBUFFER_EXT, self.id)  # Rendering off-screen
        gl.glViewport(0, 0, self.texture.width, self.texture.height)

    def unbind(self):
        """Unbind the FBO."""
        # Unbind the FBO
        gl.glBindFramebufferEXT(gl.GL_FRAMEBUFFER_EXT, 0)

        # Restore the old viewport size
        gl.glViewport(*self._old_viewport_size)


__author__ = 'ratcave'

import numpy as np
import _transformations as transformations


def calculate_model_matrix(position, rotation, scale):
    """Returns 4x4 model matrix from the (x.y,z) position and XYZ Euler rotation, in degrees."""

    # Set Model and Normal Matrices
    trans_mat = transformations.translation_matrix(position)

    rotation = tuple(np.radians(coord) for coord in rotation)
    rot_mat = transformations.euler_matrix(rotation[0], rotation[1], rotation[2])
    scale_mat = transformations.scale_matrix(scale)

    return np.dot(np.dot(trans_mat, rot_mat), scale_mat)


def calculate_view_matrix(position, rotation):
    """TReturns 4x4 view matrix from the (x.y,z) position and XYZ Euler rotation, in degrees."""
    # Set View Matrix

    trans_mat = transformations.translation_matrix((-position[0], -position[1], -position[2]))

    rotation = tuple(-np.radians(coord) for coord in rotation)
    rot_mat = transformations.euler_matrix(rotation[0], rotation[1], rotation[2], 'rxyz')

    return np.dot(rot_mat, trans_mat)





from . import utils
from . import resources
from .camera import Camera
from .fbo import FBO
from .light import Light
from .mesh import Mesh, MeshData, Material
from .mixins import Physical
from .scene import Scene
from .shader import Shader, Uniform, UniformCollection
from .texture import Texture, TextureCube
from .wavefront import WavefrontReader

__all__ = ['Camera', 'Mesh', 'MeshData', 'Material', 'Physical', 'Scene', 'WavefrontReader', 'resources']





from os import path
import copy

import numpy as np
from .mesh import Mesh, MeshData, MeshLoader, Material
from .scene import Scene


class WavefrontReader(object):

    def __init__(self, file_name):
        """
        Reads Wavefront (.obj) files created in Blender to build ratCAVE.graphics Mesh objects.

        :param file_name: .obj file to read (assumes an accompanying .mtl file has the same base file name.)
        :type file_name: str
        :return:
        :rtype: WavefrontReader
        """
        self.file_name = file_name
        self.meshdata = {}
        self.materials = {}
        self.mesh_material_names = {}
        self.mesh_names = []

        file_name, _ = path.splitext(file_name)
        self._parse_obj(file_name + '.obj')
        self._parse_mtl(file_name + '.mtl')

    def _parse_obj(self, file_name):

        with open(file_name, 'r') as wavefront_file:
            lines = wavefront_file.readlines()
            total_lines = len(lines)

            obj_idx = 0
            list_prefixes = ['#', 'mtllib', 'o', 'v', 'vt', 'vn', 'usemtl', 's', 'f', 'l']  # TODO: Find out what 'l' is for
            single_prefixes = ['mtllib', 'o', 's', 'usemtl']
            props = {pre: [] for pre in list_prefixes}
            s_props = {pre: None for pre in single_prefixes}
            props.update(s_props)

            for line_num, line in enumerate(lines):

                name = props['o']

                # Add data to corresponding list in dictionary.
                prefix, rest = line.split(' ', 1)
                if isinstance(props[prefix], list):
                    props[prefix].append(rest[:-1])
                else:
                    props[prefix] = rest[:-1]

                # Reset property dictionary for each object
                if (prefix == 'o' and props['f']) or line_num == total_lines - 1:
                    self.mesh_material_names[name] = props['usemtl']  #
                    # Build mesh from props dictionary
                    meshdata = self._build_mesh(props)
                    self.meshdata[name] = meshdata
                    self.mesh_names.append(name)

                    props['f'] = []

    def get_mesh(self, mesh_name, **kwargs):
        """
        Returns a Mesh object for directly rendering in a scene.

        :param mesh_name:
        :param kwargs: All of Mesh's keword arguments will be applied to the Mesh, for convenient Mesh creation.
        :return: Mesh
        :rtype: Mesh
        """
        meshdata = copy.deepcopy(self.meshdata[mesh_name])

        material_name = self.mesh_material_names[mesh_name] if 'material_name' not in kwargs else kwargs['material_name']
        material = copy.deepcopy(self.materials[material_name])
        return MeshLoader(mesh_name, meshdata, material).load_mesh(**kwargs)

    def get_meshes(self, mesh_names, **kwargs):
        """Returns a dictionary of meshes, with kwargs applied to all meshes identically, as in get_mesh()"""
        return {name: self.get_mesh(name, **kwargs) for name in mesh_names}

    def get_scene(self, include=[], exclude=[]):
        """
        Return a Scene object containing the Meshes in the file.

        Args:
            include (list): mesh names to only include.
            exclude (list): mesh names to exclude.

        Returns:
            :py:class:`.Scene`
        """
        names = include if include else self.mesh_names
        meshes = [self.get_mesh(name) for name in names if name not in exclude]
        return Scene(meshes)

    def _build_mesh(self, props):

        # Handle the face indices.  Need separate arrays for each mesh object.
        tmp = [line.split(' ') for line in props['f']]
        num_coords = len(tmp[0][0].split('/'))
        tmp = [line.replace('/', ' ').split(' ') for line in props['f']]
        face_indices = np.array([int(number) for inner in tmp for number in inner]).reshape((-1, num_coords)) - 1
        inds = np.arange(face_indices.shape[0])

        # Get the float data into NumPy arrays
        vertices = np.array([float(number) for inner in props['v'] for number in inner.split(' ')]).reshape((-1, 3))
        textureUVs = np.array([float(number) for inner in props['vt'] for number in inner.split(' ')]).reshape((-1, 2))
        normals = np.array([float(number) for inner in props['vn'] for number in inner.split(' ')]).reshape((-1, 3))

        # Reorder float data, so they all match the "inds" list
        verts = vertices[face_indices[:, 0]]
        textUVs = textureUVs[face_indices[:, 1]] if np.any(textureUVs) else None
        norms = normals[face_indices[:, 2]] if np.any(normals) else None

        # Build MeshData object
        meshdata = MeshData(verts, inds, norms, textUVs)
        return meshdata

    def _parse_mtl(self, filename):

        def parse_property_line(data_line):
            prefix, value = data_line.lstrip().split(' ', 1)

            try:
                value = [float(num) for num in value.split(' ')]  # convert to list of floats
                value = value[0] if len(value) == 1 else value
            except ValueError:  # If not a sequence of numbers, but a string instead
                pass

            return prefix, value

        # parsing in 3 steps:
        # - read all data in [lines] list
        # - split each mesh section in {mtls} dict
        # - parse each mesh section and create material from it

        # STEP1: read all data in [lines] list
        try:
            with open(filename, 'r') as material_file:
                lines = [line for line in material_file]

        except IOError:
            print("Warning: No .mtl material file found for "
                  ".obj file. Using default material instead...")
            self.materials[None] = Material()
            return

        mtls = {}  # like {'Shape.005': ['Ns 96.078431', 'Ka 0.166667 0.166667 0.166667', ...], ...}
        mtl_name_buff = ''
        mtl_line_buff = []

        # STEP 2: split each mesh section in {mtls} dict
        # empty lines as newmtl section separators
        separators = [0] + [i for i, x in enumerate(lines) if x == "\n"] + [len(lines)]

        for i in range(len(separators) - 1):
            for line in lines[separators[i]:separators[i + 1]]:
                if line.startswith('newmtl'):
                    mtl_name_buff = line.strip('\n').split(' ')[1]

                elif len(line.strip('\n')) > 0:
                    mtl_line_buff.append(line.strip('\n'))

            if mtl_name_buff:
                mtls[mtl_name_buff] = mtl_line_buff

            mtl_name_buff = ''
            mtl_line_buff = []

        # STEP 3: parse each newmtl section into Material
        prefixes = ['#', 'newmtl', 'Ns', 'Ka', 'Kd', 'Ks', 'Ni', 'd', 'illum', 'map_Kd']

        for name, lines in mtls.items():
            props = dict.fromkeys(prefixes, None)

            for line in lines:
                p, val = parse_property_line(line)
                if p in prefixes:
                    props[p] = val

            kwargs = {
                "diffuse": props['Kd'],
                "spec_weight": props['Ns'],
                "specular": props['Ks'],
                "ambient": props['Ka'],
                "opacity": props['d'],
                "texture_file": props['map_Kd']
            }
            self.materials[name] = Material(**kwargs)





import warnings
import pyglet.gl as gl

from . import mixins, Camera, Light, resources, mesh
from .utils import gl as glutils
from .texture import TextureCube


class Scene(object):

    def __init__(self, meshes=(), camera=None, light=None, bgColor=(0.4, 0.4, 0.4)):
        """Returns a Scene object.  Scenes manage rendering of Meshes, Lights, and Cameras."""
        # TODO: provide help to make camera aspect and fov_y for cubemapped scenes!
        # Initialize List of all Meshes to draw

        self.root = mesh.EmptyMesh()
        self.root.add_children(meshes)
        self.camera = Camera() if not camera else camera # create a default Camera object
        self.light = Light() if not light else light
        self.bgColor = bgColor

    def clear(self):
        """Clear Screen and Apply Background Color"""
        gl.glClearColor(*(self.bgColor + (1.,)))
        gl.glClear(gl.GL_COLOR_BUFFER_BIT | gl.GL_DEPTH_BUFFER_BIT)

    def draw(self, shader=resources.genShader, clear=True,
             send_mesh_uniforms=True, send_camera_uniforms=True, send_light_uniforms=True, userdata={},
             gl_states=(gl.GL_DEPTH_TEST, gl.GL_POINT_SMOOTH, gl.GL_TEXTURE_CUBE_MAP, gl.GL_TEXTURE_2D)):#, gl.GL_BLEND)):
        """Draw each visible mesh in the scene from the perspective of the scene's camera and lit by its light."""

        # Enable 3D OpenGL states (glEnable, then later glDisable)
        with glutils.enable_states(gl_states):
            # gl.glBlendFunc(gl.GL_SRC_ALPHA, gl.GL_ONE_MINUS_SRC_ALPHA)

            # Bind Shader
            with shader:

                if clear:
                    self.clear()

                # Send Uniforms that are constant across meshes.
                if send_camera_uniforms:
                    self.camera.update()
                    shader.uniform_matrixf('view_matrix', self.camera.view_matrix.T.ravel())
                    shader.uniform_matrixf('projection_matrix', self.camera.projection_matrix.T.ravel())
                    shader.uniformf('camera_position', *self.camera.position)

                if send_light_uniforms:
                    shader.uniformf('light_position', *self.light.position)

                for mesh in self.root:
                    mesh._draw(shader=shader, send_uniforms=send_mesh_uniforms)


    def draw360_to_texture(self, cubetexture, shader=resources.genShader, autoclear=True, userdata={},
             gl_states=(gl.GL_DEPTH_TEST, gl.GL_POINT_SMOOTH, gl.GL_TEXTURE_CUBE_MAP, gl.GL_TEXTURE_2D)):#, gl.GL_BLEND)):
        """
        Draw each visible mesh in the scene from the perspective of the scene's camera and lit by its light, and
        applies it to each face of cubetexture, which should be currently bound to an FBO.
        """

        assert self.camera.aspect == 1. and self.camera.fov_y == 90
        assert type(cubetexture) == TextureCube, "Must render to TextureCube"

        # Enable 3D OpenGL states (glEnable, then later glDisable)
        with glutils.enable_states(gl_states):
            # gl.glBlendFunc(gl.GL_SRC_ALPHA, gl.GL_ONE_MINUS_SRC_ALPHA)

            # Bind Shader
            with shader:

                self.camera.update()

                shader.uniformf('light_position', *self.light.position)
                shader.uniform_matrixf('projection_matrix', self.camera.projection_matrix.T.ravel())
                shader.uniformf('camera_position', *self.camera.position)

                # Pre-Calculate all 6 view matrices
                view_matrices = []
                view_matrix_loc = shader.get_uniform_location('view_matrix')
                for rotation in [[180, 90, 0], [180, -90, 0], [90, 0, 0], [-90, 0, 0], [180, 0, 0], [0, 0, 180]]:
                    self.camera.rotation = rotation
                    self.camera.update_view_matrix()
                    view_matrices.append(self.camera.view_matrix.T.ravel())

                for mesh_idx, mesh in enumerate(self.root):
                    for face, view_matrix in enumerate(view_matrices):
                        cubetexture.attach_to_fbo(face)
                        if autoclear and not mesh_idx:
                            self.clear()
                        shader.uniform_matrixf('view_matrix', view_matrix, loc=view_matrix_loc)
                        mesh._draw(shader=shader, send_uniforms=not face)





import unittest
from ratcave import resources, WavefrontReader


class TestMesh(unittest.TestCase):
    """
    Run tests from main project folder.
    """

    def setUp(self):
        self.reader = WavefrontReader(resources.obj_primitives)
        self.mesh = self.reader.get_mesh("Cube")

    def tearDown(self):
        pass

    def test_position_update(self):

        mesh = self.mesh
        for pos in [(4,5, 6), (5, 4, 1)]:
            mesh.x, mesh.y, mesh.z = pos
            self.assertEqual(mesh.position, pos)

    def test_rotation_update(self):

        mesh = self.mesh
        for rot in [(4, 5, 6), (5, 4, 1)]:
            mesh.rot_x, mesh.rot_y, mesh.rot_z = rot
            self.assertEqual(mesh.rotation, rot)





"""
    mesh
    ~~~~

    This module contains the Mesh, MeshData, and Material classes.
    This documentation was auto-generated from the mesh.py file.
"""
import copy
import numpy as np
from pyglet import gl
from .utils import gl as ugl
from . import mixins, shader
from . import texture as texture_module


class MeshData(object):

    def __init__(self, vertices, face_indices, normals, texcoords=None):
        """
        Collects all vertex data for rendering in 3D graphics packages.

        Args:
            vertices (list): Nx3 vertex array
            face_indices (list): Nx3 Face index array (0-indexed)
            normals (list): Nx3 normals array
            texture_uv (list): Nx2 texture_uv array

        Returns:
            MeshData object
        """
        # CPU Data
        self.vertices = np.array(vertices, dtype=float).reshape((-1, 3))
        self.face_indices = np.array(face_indices, dtype=np.uint16).reshape((-1, 1))
        self.normals = np.array(normals, dtype=float).reshape((-1, 3))
        self.texcoords = np.array(texcoords, dtype=float).reshape((-1, 2))
        assert self.vertices.shape[0] == self.normals.shape[0]
        assert self.vertices.shape[0] == self.texcoords.shape[0]

        self.is_loaded = False
        self.glbuffer = False

    def load(self):
        self.reindex()
        self.glbuffer = ugl.VAO(indices=self.face_indices)
        with self.glbuffer:
            for loc, verts in enumerate([self.vertices, self.normals, self.texcoords]):
                self.glbuffer.assign_vertex_attrib_location(ugl.VBO(verts), loc)
        self.is_loaded = True

    def draw(self, mode):
        if not self.is_loaded:
            self.load()

        with self.glbuffer as vao:
            vao.draw(mode)

    def reindex(self):

        def to_joined_struct_array_view(*arrays):
            """Concatenates (columnwise) each array given as input, then returns it as a structured array."""
            array = np.hstack(arrays)
            dtype = array.dtype.descr * array.shape[1]
            return array.view(dtype)

        all_vert_combs = to_joined_struct_array_view(self.vertices, self.normals, self.texcoords)
        unique_combs = np.unique(all_vert_combs)
        unique_vert_combs_sorted = np.sort(unique_combs)

        self.face_indices = np.array([np.searchsorted(unique_vert_combs_sorted, vert) for vert in all_vert_combs],
                                     dtype=np.uint32).reshape((-1, 1))
        unique_combs_array = unique_combs.view(float).reshape((unique_combs.shape[0], -1))
        self.vertices = unique_combs_array[:, :3]
        self.normals = unique_combs_array[:, 3:6]
        self.texcoords = unique_combs_array[:, 6:]


class Material(object):

    def __init__(self, diffuse=[.8, .8, .8], spec_weight=0., specular=[0., 0., 0.],
                 ambient=[0., 0., 0.], opacity=1., flat_shading=False, texture_file=None):
        self.diffuse = diffuse
        self.spec_weight = spec_weight
        self.specular = specular
        self.ambient = ambient
        self.opacity = opacity
        self.flat_shading = flat_shading
        self.texture_file = texture_file


class MeshLoader(object):

    def __init__(self, name, meshdata, material=None):
        """Creates various types of Meshes from MeshData and Material objects."""

        self.name = name
        self.meshdata = meshdata
        self.material = material

    def load_mesh(self, **kwargs):
        from collections import Iterable

        """Construct a Mesh object"""
        uniforms = shader.UniformCollection()
        texture = None

        if self.material:
            for key, val in list(self.material.__dict__.items()):
                if key == "texture_file":
                    if val is not None:
                        texture = texture_module.Texture.from_image(val)
                else:
                    newval = copy.deepcopy(val)
                    if not isinstance(newval, Iterable):
                        newval = int(val) if isinstance(val, bool) else newval
                        newval = [newval]

                    uniforms[key] = shader.Uniform(key, *newval)

        return Mesh(self.name, self.meshdata, uniforms=uniforms, texture=texture, **kwargs)


class EmptyMesh(mixins.PhysicalNode):

    def __init__(self, *args, **kwargs):
        super(EmptyMesh, self).__init__(*args, **kwargs)

    def _draw(self, shader=None, **kwargs):
        pass


class Mesh(EmptyMesh, mixins.Picklable):

    drawstyle = {'fill': gl.GL_TRIANGLES, 'line': gl.GL_LINE_LOOP, 'point': gl.GL_POINTS}

    def __init__(self, name, meshdata, uniforms=shader.UniformCollection(), drawstyle='fill', visible=True,
                 point_size=4, texture=None, **kwargs):
        """
        Returns a Mesh object, containing the position, rotation, and color info of an OpenGL Mesh.

        Meshes have two coordinate system, the "local" and "world" systems, on which the transforms are performed
        sequentially.  This allows them to be placed in the scene while maintaining a relative position to one another.

        .. note:: Meshes are not usually instantiated directly, but from a 3D file, like the WavefrontReader .obj and .mtl files.

        Args:
            name (str): the mesh's name.
            vertices: the Nx3 vertex coordinate data
            normals: the Nx3 normal coordinate data
            texcoords: the Nx2 texture coordinate data
            uniforms (list): a list of all Uniform objects
            drawstyle (str): 'point': only vertices, 'line': points and edges, 'fill': points, edges, and faces (full)
            visible (bool): whether the Mesh is available to be rendered.  To make hidden (invisible), set to False.
            point_size (int): How big to draw the points, when drawstyle is 'point'

        Returns:
            Mesh instance
        """
        self.data = meshdata

        super(Mesh, self).__init__(**kwargs)

        self.name = name

        # Convert Mean position into Global Coordinates. If "centered" is True, though, simply leave global position to 0
        vertex_mean = np.mean(self.data.vertices, axis=0)
        self.data.vertices -= vertex_mean
        self.position = vertex_mean if 'position' not in kwargs else kwargs['position']

        #: :py:class:`.Physical`, World Mesh coordinates
        #: Local Mesh coordinates (Physical type)
        self.uniforms = uniforms if type(uniforms) == shader.UniformCollection else shader.UniformCollection(uniforms)

        #: Pyglet texture object for mapping an image file to the vertices (set using Mesh.load_texture())
        self.texture = texture or texture_module.BaseTexture()
        self.drawstyle = drawstyle
        self.point_size = point_size

        #: Bool: if the Mesh is visible for rendering. If false, will not be rendered.
        self.visible = visible
        self.vao = None

        self.is_updated = False

    def __str__(self):
        return "%s(%d) at %s" % (self.name, len(self.children), self.position_global)

    def __repr__(self):
        return str(self)

    def update(self):
        super(Mesh, self).update()

        vertices_local = np.vstack([self.data.vertices.T, np.ones(len(self.data.vertices))])
        vertices_global = np.dot(self.model_matrix_global, vertices_local).T

        # if needed, one can store global vertex coordinates
        vg = vertices_global

        self.min_xyz = np.array((vg[:, 0].min(), vg[:, 1].min(), vg[:, 2].min()))
        self.max_xyz = np.array((vg[:, 0].max(), vg[:, 1].max(), vg[:, 2].max()))

    def _draw(self, shader=None, send_uniforms=True, *args, **kwargs):
        super(Mesh, self)._draw(*args, **kwargs)

        if not self.is_updated:
            self.update()
            self.is_updated = True

        if self.visible:

            # Bind the VAO and Texture, and draw.
            with self.texture as texture:

                # Change Material to Mesh's
                if send_uniforms:

                    self.update_model_and_normal_matrix()
                    self.uniforms.send_to(shader)
                    texture.uniforms.send_to(shader)

                    # Send Model and Normal Matrix to shader.
                    try:
                        shader.uniform_matrixf('model_matrix', self.model_matrix_global.T.ravel(), loc=self.modelmat_loc)
                        shader.uniform_matrixf('normal_matrix', self.normal_matrix_global.T.ravel(), loc=self.normalmat_loc)
                    except AttributeError:
                        self.modelmat_loc = shader.get_uniform_location('model_matrix')
                        self.normalmat_loc = shader.get_uniform_location('normal_matrix')
                        pass

                # Set Point Size, if drawing a point cloud
                if self.drawstyle == 'point':
                    gl.glPointSize(int(self.point_size))

                # Send in the vertex and normal data
                self.data.draw(Mesh.drawstyle[self.drawstyle])



__author__ = 'ratcave'

from setuptools import setup, find_packages, Extension
import numpy


setup(name='ratcave',
      version='0.5post2',
      url='https://www.github.com/neuroneuro15/ratcave',
      description='3D Graphics Engine for CAVE VR setups',
      author='Nicholas A. Del Grosso',
      author_email='delgrosso.nick@gmail.com',
      license='MIT',
      packages=find_packages(exclude=['docs']),
      include_package_data=True,
      package_data={'': ['../assets/*.'+el for el in ['png', 'obj', 'mtl']] +
                        ['../shaders/*'+el for el in ['vert', 'frag']]
                    },
      install_requires=['pyglet', 'numpy'],
      ext_modules=[Extension('_transformations', sources=['c_sources/transformations.c'], include_dirs=[numpy.get_include()])],
      test_suite='tests',
      keywords='graphics 3D pyglet psychopy python virtual reality VR',
      classifiers=[
          "Topic :: Multimedia :: Graphics :: 3D Rendering",
          "Development Status :: 3 - Alpha",
          "License :: OSI Approved :: MIT License",
          "Intended Audience :: Science/Research",
          "Programming Language :: Python",
          "Programming Language :: Python :: 2",
          "Programming Language :: Python :: 2.7",
          "Programming Language :: Python :: 3",
          "Programming Language :: Python :: 3.3",
          "Programming Language :: Python :: 3.4",
          "Programming Language :: Python :: 3.5",
          "Programming Language :: Python :: 3.6",
      ],
      )




from.utils import gl as ugl
import pyglet
import pyglet.gl as gl
from . import shader


class BaseTexture(object):

    int_flag = 0
    tex_name = 'TextureMap'
    cube_name = 'CubeMap'

    def __init__(self):
        self.uniforms = shader.UniformCollection({
            self.tex_name: shader.Uniform(self.tex_name, 0),
            self.cube_name: shader.Uniform(self.cube_name, 0),
            'textype': shader.Uniform('textype', self.int_flag),
        })

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        pass


class Texture(BaseTexture, ugl.BindTargetMixin):

    target = gl.GL_TEXTURE_2D
    target0 = gl.GL_TEXTURE_2D
    attachment_point = gl.GL_COLOR_ATTACHMENT0_EXT
    internal_fmt = gl.GL_RGBA
    pixel_fmt=gl.GL_RGBA
    _all_slots = list(range(1, 25))[::-1]
    int_flag = 1
    bindfun = gl.glBindTexture

    def __init__(self, id=None, width=1024, height=1024, data=None):
        """2D Color Texture class. Width and height can be set, and will generate a new OpenGL texture if no id is given."""

        self._slot = self._all_slots.pop()
        self.uniforms = shader.UniformCollection({
            self.tex_name: shader.Uniform(self.tex_name, self._slot),
            self.cube_name: shader.Uniform(self.cube_name, 0),
            'textype': shader.Uniform('textype', self.int_flag),
        })

        if id != None:
            self.id = id
            self.data = data  # This is used for anything that might be garbage collected (i.e. pyglet textures)
        else:
            self.id = ugl.create_opengl_object(gl.glGenTextures)
            self.width = width
            self.height = height
            self.bind()
            self._apply_filter_settings()
            self._genTex2D()

        self.unbind()

    @property
    def slot(self):
        """The texture's ActiveTexture slot."""
        return self._slot

    def __enter__(self):
        gl.glActiveTexture(gl.GL_TEXTURE0 + self.slot)
        self.bind()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.unbind()
        gl.glActiveTexture(gl.GL_TEXTURE0)

    @staticmethod
    def _generate_id():
        return ugl.create_opengl_object(gl.glGenTextures)

    def _genTex2D(self):
        """Creates an empty texture in OpenGL."""
        gl.glTexImage2D(self.target0, 0, self.internal_fmt, self.width, self.height, 0, self.pixel_fmt, gl.GL_UNSIGNED_BYTE, 0)

    def _apply_filter_settings(self):
        """Applies some hard-coded texture filtering settings."""
        # TODO: Allow easy customization of filters
        gl.glTexParameterf(self.target, gl.GL_TEXTURE_MIN_FILTER, gl.GL_LINEAR)
        gl.glTexParameterf(self.target, gl.GL_TEXTURE_MAG_FILTER, gl.GL_LINEAR)
        gl.glTexParameterf(self.target, gl.GL_TEXTURE_WRAP_S, gl.GL_CLAMP_TO_EDGE)
        gl.glTexParameterf(self.target, gl.GL_TEXTURE_WRAP_T, gl.GL_CLAMP_TO_EDGE)

    def attach_to_fbo(self):
        """Attach the texture to a bound FBO object, for rendering to texture."""
        gl.glFramebufferTexture2DEXT(gl.GL_FRAMEBUFFER_EXT, self.attachment_point, self.target0, self.id, 0)

    @classmethod
    def from_image(cls, img_filename, **kwargs):
        """Uses Pyglet's image.load function to generate a Texture from an image file."""
        img = pyglet.image.load(img_filename)
        tex = img.get_texture()
        gl.glBindTexture(gl.GL_TEXTURE_2D, 0)
        return cls(id=tex.id, data=tex, **kwargs)


class TextureCube(Texture):

    target = gl.GL_TEXTURE_CUBE_MAP
    target0 = gl.GL_TEXTURE_CUBE_MAP_POSITIVE_X
    int_flag = 2

    def __init__(self, *args, **kwargs):
        """the Color Cube Texture class."""
        # TODO: check that width == height!
        super(TextureCube, self).__init__(*args, **kwargs)

        self.uniforms = shader.UniformCollection({
            self.tex_name: shader.Uniform(self.tex_name, 0),
            self.cube_name: shader.Uniform(self.cube_name, self._slot),
            'textype': shader.Uniform('textype', self.int_flag),
        })

    def _apply_filter_settings(self, *args, **kwargs):
        super(TextureCube, self)._apply_filter_settings()
        with self:
            gl.glTexParameterf(self.target, gl.GL_TEXTURE_WRAP_R, gl.GL_CLAMP_TO_EDGE)

    def _genTex2D(self):
        """Generate an empty texture in OpenGL"""
        for face in range(6):
            gl.glTexImage2D(self.target0 + face, 0, self.internal_fmt, self.width, self.height, 0,
                            self.pixel_fmt, gl.GL_UNSIGNED_BYTE, 0)

    @classmethod
    def from_image(cls, img_filename):
        raise NotImplementedError()

    def attach_to_fbo(self, face=0):
        gl.glFramebufferTexture2DEXT(gl.GL_FRAMEBUFFER_EXT, self.attachment_point,
                                     self.target0 + face,
                                     self.id,  0)  # Select face of cube texture to render to.

class DepthTexture(Texture):
    internal_fmt = gl.GL_DEPTH_COMPONENT
    pixel_fmt = gl.GL_DEPTH_COMPONENT
    attachment_point = gl.GL_DEPTH_ATTACHMENT_EXT


class GrayscaleTexture(Texture):
    internal_fmt = gl.GL_R8
    pixel_fmt = gl.GL_RED


class GrayscaleTextureCube(TextureCube):
    internal_fmt = gl.GL_R8
    pixel_fmt = gl.GL_RED


class RenderBuffer(ugl.BindingContextMixin, ugl.BindTargetMixin):

    target = gl.GL_RENDERBUFFER_EXT
    attachment_point = gl.GL_DEPTH_ATTACHMENT
    internal_fmt = gl.GL_DEPTH_COMPONENT24
    bindfun = gl.glBindRenderbufferEXT

    def __init__(self, width, height):

        self.id = ugl.create_opengl_object(gl.glGenRenderbuffersEXT)
        self.width = width
        self.height = height
        self.bind()
        self._gen()

    def _gen(self):
        gl.glRenderbufferStorageEXT(self.target, self.internal_fmt, self.width, self.height)

    def attach_to_fbo(self):
        gl.glFramebufferRenderbufferEXT(gl.GL_FRAMEBUFFER_EXT, self.attachment_point, self.target, self.id)




import numpy as np
import pickle
from . import utils
from collections import deque

# TODO: Check for loops and duplicate nodes in the Scene graph
class SceneNode(object):

    def __init__(self, parent=None, children=None):
        """A Node of the Scenegraph.  Has children, but no parent."""
        self._children = []
        self._parent = None
        if parent:
            self.parent = parent
        if children:
            self.add_children(children)

    def __iter__(self):
        """Returns an iterator that walks through the scene graph,
         starting with the current object."""
        def walk_tree_breadthfirst(obj):
            """tree walking algorithm, using algorithm from
            http://kmkeen.com/python-trees/"""
            order = deque([obj])
            while len(order) > 0:
                order.extend(order[0]._children)
                yield order.popleft()

        return walk_tree_breadthfirst(self)

    @property
    def parent(self):
        """A SceneNode object that is this object's parent in the scene graph."""
        return self._parent

    @parent.setter
    def parent(self, value):
        assert isinstance(value, SceneNode)
        if self._parent is not None:
            self._parent._children.remove(self)
        self._parent = value
        self._parent._children.append(self)

    def add_children(self, children=list()):
        """Adds a list of objects as children in the scene graph."""

        for child in children:
            assert isinstance(child, SceneNode)
            child._parent = self
            self._children.append(child)

    def remove_children(self, children):
        for child in children:
            child._parent = None
            self._children.remove(child)

    @property
    def children(self):
        return tuple(self._children)


class Physical(object):

    __observables = ('x', 'y', 'z', 'rot_x', 'rot_y', 'rot_z', 'scale')

    def __init__(self, position=(0., 0., 0.), rotation=(0., 0., 0.), scale=1., *args, **kwargs):
        """XYZ Position, Scale and XYZEuler Rotation Class.

        Args:
            position (list): (x, y, z) translation values.
            rotation (list): (x, y, z) rotation values
            scale (float): uniform scale factor. 1 = no scaling.
        """
        super(Physical, self).__init__(*args, **kwargs)
        self.__dict__['x'], self.__dict__['y'], self.__dict__['z'] = position
        self.__dict__['rot_x'], self.__dict__['rot_y'], self.__dict__['rot_z'] = rotation
        self.__dict__['scale'] = scale
        self._rot_matrix = None

        self.model_matrix = np.zeros((4,4))
        self.normal_matrix = np.zeros((4,4))
        self.view_matrix = np.zeros((4,4))

        # rectangular boundaries
        self.min_xyz = np.array((0, 0, 0))
        self.max_xyz = np.array((0, 0, 0))

        self.__oldinfo = tuple()
        self.update()

    def __setattr__(self, key, value):
        super(Physical, self).__setattr__(key, value)

        if key in Physical.__observables:
            self.on_change()

    def on_change(self):
        """
        This method fires when object position or geometry changes.
        Can be overwritten by parent classes to add more actions.
        """
        self.update()

    @property
    def position(self):
        """xyz local position"""
        return self.x, self.y, self.z

    @position.setter
    def position(self, value):
        self.x, self.y, self.z = value

    @property
    def rotation(self):
        """XYZ Euler rotation, in degrees"""
        return self.rot_x, self.rot_y, self.rot_z

    @rotation.setter
    def rotation(self, value):
        self.rot_x, self.rot_y, self.rot_z = value

    def __gen_info_summary(self):
        return  self.position + self.rotation + (self.scale,)

    def has_changed(self):
        return self.__oldinfo != self.__gen_info_summary()

    def update_model_matrix(self):
        self.model_matrix = utils.orienting.calculate_model_matrix(self.position, self.rotation, self.scale)

    def update_model_and_normal_matrix(self):
        if self.has_changed():
            self.update_model_matrix()
            self.normal_matrix = np.linalg.inv(self.model_matrix.T)  # which order is correct: t then i, or the reverse?

    def update_view_matrix(self):
        # if self.has_changed():  # TODO: Getting some bugs with this line, not sure why
        self.view_matrix = utils.orienting.calculate_view_matrix(self.position, self.rotation)

    def update(self):
        """Calculate model, normal, and view matrices from position, rotation, and scale data."""
        if self.has_changed():
            self.update_model_and_normal_matrix()
            self.update_view_matrix()
            self.__oldinfo = self.__gen_info_summary()


class PhysicalNode(Physical, SceneNode):

    def __init__(self, *args, **kwargs):
        """Object with xyz position and rotation properties that are relative to its parent."""
        self.model_matrix_global = np.zeros((4,4))
        self.normal_matrix_global = np.zeros((4,4))
        self.view_matrix_global = np.zeros((4,4))
        super(PhysicalNode, self).__init__(*args, **kwargs)

    def update(self):
        super(PhysicalNode, self).update()

        """Calculate world matrix values from the dot product of the parent."""
        if self.parent:
            self.model_matrix_global = np.dot(self.parent.model_matrix_global, self.model_matrix)
            self.normal_matrix_global = np.dot(self.parent.normal_matrix_global, self.normal_matrix)
            self.view_matrix_global = np.dot(self.parent.normal_matrix_global, self.normal_matrix)
        else:
            self.model_matrix_global = self.model_matrix
            self.normal_matrix_global = self.normal_matrix
            self.view_matrix_global = self.view_matrix

    @property
    def position_global(self):
        return tuple(self.model_matrix_global[:3, -1].tolist())


class Picklable(object):

    def save(self, filename):
        """Save the object to a file.  Will be Pickled in the process, but can be loaded easily with Class.load()"""
        with open(filename, 'wb') as f:
            pickle.dump(self, f)

    @classmethod
    def load(cls, filename):
        """Load the object from a pickle file."""
        with open(filename) as f:
            obj = pickle.load(f)
            assert isinstance(obj, cls), "File's object is {}, but should be {}".format(type(obj), cls)
        return obj
